/* 
    buz - Iomega Buz frame grabber driver

    Copyright (C) 1998 Dave Perks <dperks@ibm.net>

	based on

    bttv - Bt848 frame grabber driver

    Copyright (C) 1996,97,98 Ralph  Metzler (rjkm@thp.uni-koeln.de)
                           & Marcus Metzler (mocm@thp.uni-koeln.de)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/wrapper.h>

#include <linux/videodev.h>
#include <linux/version.h>
#include <asm/uaccess.h>

#include "/usr/src/linux/drivers/char/i2c.h"
#include "buz.h"
#include "video_decoder.h"

/*
GPIO0 = 1, take board out of reset
GPIO1 = 1, take JPEG codec out of sleep mode
GPIO3 = 1, deassert FRAME# to 36060

SAA7111A mode 0 - Composite input
	mode 1 - S-Video input?
	mode 2 - noise
	mode 3 - Composite
	mode 4 - nothing
	mode 5 - S-Video
	mode 6 - nothing
	mode 7 - both
*/

#undef KERN_DEBUG
#define KERN_DEBUG

#define DEBUG(x)	x	/* Debug driver */	
#define IDEBUG(x) 	x	/* Debug interrupt handler */


static unsigned int vidmem=0;   /* manually set video mem address */
static int triton1=0;

static int input=0;
static int gpio=0xd7;

MODULE_PARM(vidmem,"i");
MODULE_PARM(triton1,"i");
MODULE_PARM(input,"i");
MODULE_PARM(gpio,"i");

static void * rvmalloc(unsigned long size);
static void rvfree(void * mem, unsigned long size);
static int fbuffer_alloc(struct buz *btv);
static void fbuffer_build_frag_tables(struct buz *btv);

static void i2c_setlines(struct i2c_bus *bus,int ctrl,int data);
static int i2c_getdataline(struct i2c_bus *bus);

static void GPIO(struct buz *btv, unsigned bit, unsigned value) ;

static void zr36057_dump_range(struct buz * btv, unsigned from, unsigned to);
static void zr36057_dump(struct buz * btv);
static void zr36057_cap(struct buz *btv, uint state);
static void zr36057_set_vfe(struct buz *btv);
static void zr36057_set_jpg(struct buz *btv);
static void zr36057_enable_jpg(struct buz *btv, int enable);
static void zr36057_set_winsize(struct buz *btv);
static void write_overlay_mask(struct buz *btv, struct video_clip *vp, int count);

static int vgrab(struct buz *btv, struct video_mmap *mp);
static int vsync(struct buz * btv, int frame);

static int find_vga(void);
static void handle_chipset(void);

static void buz_reap_stat_com(struct buz * btv);
static void buz_irq(int irq, void *dev_id, struct pt_regs * regs);

static int buz_open(struct video_device *dev, int flags);
static void buz_close(struct video_device *dev);
static long buz_write(struct video_device *v, const char *buf, unsigned long count, int nonblock);
static long buz_read(struct video_device *v, char *buf, unsigned long count, int nonblock);
static int buz_ioctl(struct video_device *dev, unsigned int cmd, void *arg);
static int buz_mmap(struct video_device *dev, const char *adr, unsigned long size);
static int buz_init_done(struct video_device *dev);

static void new_risc_clip(struct video_window *vw, struct video_clip *vcp, int x, int y, int w, int h);

static int post_office_wait(struct buz *btv);
static int post_office_write(struct buz *btv, unsigned guest, unsigned reg, unsigned value);
static int post_office_read(struct buz *btv, unsigned guest, unsigned reg);

static int zr36060_write_8(struct buz *btv, unsigned reg, unsigned val);
static int zr36060_write_16(struct buz *btv, unsigned reg, unsigned val);
static int zr36060_write_24(struct buz *btv, unsigned reg, unsigned val);
static int zr36060_write_32(struct buz *btv, unsigned reg, unsigned val);
static u32 zr36060_read_8(struct buz *btv, unsigned reg);
static u32 zr36060_read_16(struct buz *btv, unsigned reg);
static u32 zr36060_read_24(struct buz *btv, unsigned reg);
static u32 zr36060_read_32(struct buz *btv, unsigned reg);
static int zr36060_start(struct buz *btv);
static int zr36060_reset(struct buz *btv);
static void zr36060_sleep(struct buz *btv, int sleep);
static int zr36060_go(struct buz *btv);
static void zr36060_dump(struct buz * btv);
static void zr36060_set_jpg(struct buz * btv);
static void zr36060_set_video(struct buz * btv);
static void zr36060_set_jpg_SOF(struct buz * btv);
static void zr36060_set_jpg_SOS(struct buz * btv);
static void zr36060_set_jpg_DRI(struct buz * btv);
static void zr36060_set_jpg_DQT(struct buz * btv);
static void zr36060_set_jpg_DHT(struct buz * btv);
static void zr36060_set_jpg_APP(struct buz * btv);
static void zr36060_set_jpg_COM(struct buz * btv);
static void zr36060_set_cap(struct buz * btv);

static void video_reset(struct buz * btv);
static int zr36057_init(int i);
static int find_zr36057(void);
static void release_buz(void);

/* Anybody who uses more than four? */
#define BUZ_MAX 4

#define	BUZ_MAX_WIDTH	720		/* never display more than 720 pixels */
#define	BUZ_MAX_HEIGHT	576		/* never display more than 576 rows */
#define	BUZ_MIN_WIDTH	32		/* never display less than 32 pixels */
#define	BUZ_MIN_HEIGHT	32		/* never display less than 32 rows */

static int buz_num;			/* number of Bt848s in use */
static struct buz buzs[BUZ_MAX];

#define I2C_DELAY   10



/*******************************/
/* Memory management functions */
/*******************************/

/* convert virtual user memory address to physical address */
/* (virt_to_phys only works for kmalloced kernel memory) */

static inline unsigned long uvirt_to_phys(unsigned long adr)
{
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *ptep, pte;

	pgd = pgd_offset(current->mm, adr);
	if (pgd_none(*pgd))
		return 0;
	pmd = pmd_offset(pgd, adr);
	if (pmd_none(*pmd))
		return 0;
	ptep = pte_offset(pmd, adr/*&(~PGDIR_MASK)*/);
	pte = *ptep;
	if(pte_present(pte))
		return 
		  virt_to_phys((void *)(pte_page(pte)|(adr&(PAGE_SIZE-1))));
	return 0;
}

static inline unsigned long uvirt_to_bus(unsigned long adr) 
{
	/*  printk("adr: 0x%8x, ",adr);
	printk("phys: 0x%8x, ",(uvirt_to_phys(adr)));
	printk("bus: 0x%8x\n",virt_to_bus(phys_to_virt(uvirt_to_phys(adr))));
	*/
	return virt_to_bus(phys_to_virt(uvirt_to_phys(adr)));
}

/* convert virtual kernel memory address to physical address */
/* (virt_to_phys only works for kmalloced kernel memory) */

static inline unsigned long kvirt_to_phys(unsigned long adr) 
{
	return uvirt_to_phys(VMALLOC_VMADDR(adr));
}

static inline unsigned long kvirt_to_bus(unsigned long adr) 
{
	return uvirt_to_bus(VMALLOC_VMADDR(adr));
}

static void * rvmalloc(unsigned long size)
{
	void * mem;
	unsigned long adr, page;

	mem=vmalloc(size);
	if (mem) 
	{
	        adr=(unsigned long) mem;
		while (size > 0) 
                {
	                page = kvirt_to_phys(adr);
			mem_map_reserve(MAP_NR(phys_to_virt(page)));
			adr+=PAGE_SIZE;
			size-=PAGE_SIZE;
		}
	}
	return mem;
}

static void rvfree(void * mem, unsigned long size)
{
        unsigned long adr, page;

	if (mem) 
	{
	        adr=(unsigned long) mem;
		while (size > 0) 
                {
	                page = kvirt_to_phys(adr);
			mem_map_unreserve(MAP_NR(phys_to_virt(page)));
			adr+=PAGE_SIZE;
			size-=PAGE_SIZE;
		}
		vfree(mem);
	}
}

/*
 *	Create the giant waste of buffer space we need for now
 *	until we get DMA to user space sorted out (probably 2.3.x)
 *
 *	We only create this as and when someone uses mmap
 */

static int fbuffer_alloc(struct buz *btv)
{
	if(!btv->fbuffer)
		btv->fbuffer=(unsigned char *) rvmalloc(BUZ_NUM_FBUF*BUZ_MAX_FBUF);
	else
		printk(KERN_ERR "buz: Double alloc of fbuffer!\n");
	if(!btv->fbuffer)
		return -ENOBUFS;
	fbuffer_build_frag_tables(btv);
	return 0;
}

static void fbuffer_build_frag_tables(struct buz *btv)
{
	int i;
	u32 vadr;
	u32 * frag_tab;
	u32 bl;
	u32 todo;

	/* build fragment tables now */
	for (i = 0; i < BUZ_NUM_FBUF; ++i)
	{
		vadr = (u32)btv->fbuffer + BUZ_MAX_FBUF * i;
		todo = BUZ_MAX_FBUF;
		frag_tab = btv->frag_tab[i];

		/* write frag tab entries for each page or part thereof */
		while (todo)
		{
			/* touch each page so there's real memory there so we don't page fault during mmap */
			*(u32 *)vadr = 0;
			bl = PAGE_SIZE  - ((PAGE_SIZE - 1) & vadr);
			if (bl > todo)
				bl = todo;
			frag_tab[0] = kvirt_to_bus(vadr);
			frag_tab[1] = (bl / 4) << 1;
			frag_tab += 2;
			vadr += bl;
			todo -= bl;
		}
		frag_tab[1-2] |= 1;	/* add Final flag */
		printk(KERN_DEBUG "buz: frag_tab[%d]=%p used_size=0x%x\n", i, btv->frag_tab[i], (u32)frag_tab - (u32)btv->frag_tab[i]);
	}
}


/* ----------------------------------------------------------------------- */

/* I2C functions                                                           */


/* software I2C functions */

static void i2c_setlines(struct i2c_bus *bus,int ctrl,int data)
{
        struct buz *btv = (struct buz*)bus->data;
	btwrite((data<<1)|ctrl, ZR36057_I2CBR);
	udelay(I2C_DELAY);
}

static int i2c_getdataline(struct i2c_bus *bus)
{
        struct buz *btv = (struct buz*)bus->data;
	return (btread(ZR36057_I2CBR)>>1)&1;
}

void attach_inform(struct i2c_bus *bus, int id)
{
	printk(KERN_INFO "buz: i2c attach %02x\n", id);
}

void detach_inform(struct i2c_bus *bus, int id)
{
	printk(KERN_INFO "buz: i2c detach %02x\n", id);
}

static struct i2c_bus buz_i2c_bus_template = 
{
        "zr36057",
        I2C_BUSID_BT848,
	NULL,

	SPIN_LOCK_UNLOCKED,

	attach_inform,
	detach_inform,

	i2c_setlines,
	i2c_getdataline,
	NULL,
	NULL,
};


/* ----------------------------------------------------------------------- */

static void GPIO(struct buz *btv, unsigned bit, unsigned value) 
{
	u32 reg;
	u32 mask;

	mask = 1 << (24 + bit);
	reg = btread(ZR36057_GPPGCR1) & ~mask;
	if (value)
		reg |= mask;
	btwrite(reg, ZR36057_GPPGCR1);
}

static void zr36057_dump_range(struct buz * btv, unsigned from, unsigned to)
{
	int num;

	num = 0;
	while (from <= to)
	{
		if (num > 3)
		{
			printk("\n");
			num = 0;
		}
		if (num == 0)
		{
			printk(KERN_INFO "buz: %03x", from);
		}
		printk(" %08x", btread(from));
		from += 4;
		num += 1;
	}
	printk("\n");
}

static void zr36057_dump(struct buz * btv)
{
	zr36057_dump_range(btv, 0x000, 0x044);
	zr36057_dump_range(btv, 0x100, 0x12c);
}

static void zr36057_cap(struct buz *btv, uint state)
{
	if (state
	&& btv->got_window
	&& btv->got_buffer)
	{
		btv->cap=1;
		btwrite(btread(ZR36057_VDCR) | ZR36057_VDCR_VidEn, ZR36057_VDCR);
		//zr36057_dump(btv);
	}
	else
	{
		btv->cap=0;
		btwrite(btread(ZR36057_VDCR) & ~ZR36057_VDCR_VidEn, ZR36057_VDCR);
	}
}


/*
 *	Set the registers for the size we have specified. Don't bother
 *	trying to understand this without the ZR36057 manual in front of
 *	you [AC]. 
 *
 *	PS: The manual is free for download in .pdf format from 
 *	www.zoran.com - nicely done those folks.
 */

struct tvnorm 
{
        u16 Wt, Wa, Ht, Ha, HStart, VStart;
};

static struct tvnorm tvnorms[] = {
	/* PAL-BDGHI */
        { 864, 720, 625, 576, 132, 10 },
	/* NTSC */
	{ 858, 720, 525, 480, 122, 10 },
	/* SECAM */
        { 864, 720, 625, 576, 132, 10 },
	/* PAL-M */
	/* PAL-N */
};
#define TVNORMS (sizeof(tvnorms)/sizeof(tvnorm))

/* set geometry
*/

static void zr36057_set_vfe(struct buz *btv)
{
	struct tvnorm *tvn;
	unsigned HStart, HEnd, VStart, VEnd;
	unsigned DispMode;
	unsigned VidWinWid, VidWinHt;
	unsigned hcrop1, hcrop2, vcrop1, vcrop2;
	unsigned We, He;
	unsigned X, Y, HorDcm, VerDcm;
	u32 reg;
	unsigned mask_line_size;

	tvn=&tvnorms[btv->win.norm];

	if (btv->win.cropwidth>tvn->Wa)
                btv->win.cropwidth=tvn->Wa;
	if (btv->win.cropheight>tvn->Ha)
	        btv->win.cropheight=tvn->Ha;

	if (btv->win.width>btv->win.cropwidth)
                btv->win.width=btv->win.cropwidth;
	if (btv->win.height>btv->win.cropheight)
	        btv->win.height=btv->win.cropheight;

	btv->win.interlace = (btv->win.height>btv->win.cropheight/2) ? 1 : 0;
	printk(KERN_INFO "buz: cropwidth=%x cropheight=%x width=%x height=%x interlace=%x\n",
		btv->win.cropwidth, btv->win.cropheight, btv->win.width, btv->win.height, btv->win.interlace);
	printk(KERN_INFO "buz: vidadr=%x x=%x y=%x bpp=%x bpl=%x\n",
		btv->win.vidadr, btv->win.x, btv->win.y, btv->win.bpp, btv->win.bpl);

	/**** zr36057 ****/

	/* horizontal */
	VidWinWid = btv->win.width;
	X = (VidWinWid * 64 + tvn->Wa - 1) / tvn->Wa;
	We = (VidWinWid * 64) / X;
	HorDcm = 64 - X;
	hcrop1 = 2 * ((tvn->Wa - We) / 4);
	hcrop2 = tvn->Wa - We - hcrop1;
	HStart = tvn->HStart;
	HEnd = HStart + tvn->Wa - 1;
	HStart += hcrop1;
	HEnd -= hcrop2;
	reg = ((HStart & ZR36057_VFEHCR_Hmask) << ZR36057_VFEHCR_HStart)
	    | ((HEnd & ZR36057_VFEHCR_Hmask) << ZR36057_VFEHCR_HEnd);
	btwrite(reg, ZR36057_VFEHCR);

	/* Vertical */
	DispMode = btv->win.interlace^1;
	VidWinHt = DispMode? btv->win.height: btv->win.height/2;
	Y = (VidWinHt * 64 * 2 + tvn->Ha - 1) / tvn->Ha;
	He = (VidWinHt * 64) / Y;
	VerDcm = 64 - Y;
	vcrop1 = (tvn->Ha / 2 - He) / 2;
	vcrop2 = tvn->Ha / 2 - He - vcrop1;
	VStart = tvn->VStart;
	VEnd = VStart + tvn->Ha/2 - 1;
	VStart += vcrop1;
	VEnd -= vcrop2;
	reg = ((VStart & ZR36057_VFEVCR_Vmask) << ZR36057_VFEVCR_VStart)
	    | ((VEnd & ZR36057_VFEVCR_Vmask) << ZR36057_VFEVCR_VEnd);
	reg |= ZR36057_VFEVCR_VSPol;
	btwrite(reg, ZR36057_VFEVCR);

	/* scaler and pixel format */
	reg = ZR36057_VFESPFR_ExtFl
	    | (HorDcm << ZR36057_VFESPFR_HorDcm)
	    | (VerDcm << ZR36057_VFESPFR_VerDcm)
	    | (DispMode << ZR36057_VFESPFR_DispMode)
	    | ZR36057_VFESPFR_LittleEndian;
reg |= ZR36057_VFESPFR_VCLKPol;
	switch (btv->win.color_fmt)
	{
	case VIDEO_PALETTE_YUV422: 
		reg |= ZR36057_VFESPFR_YUV422;
		break;
	case VIDEO_PALETTE_RGB555: 
		reg |= ZR36057_VFESPFR_RGB555 | ZR36057_VFESPFR_ErrDif;
		break;
	case VIDEO_PALETTE_RGB565: 
		reg |= ZR36057_VFESPFR_RGB565 | ZR36057_VFESPFR_ErrDif;
		break;
        case VIDEO_PALETTE_RGB24:
		reg |= ZR36057_VFESPFR_RGB888 | ZR36057_VFESPFR_Pack24;
		break;
	case VIDEO_PALETTE_RGB32:
		reg |= ZR36057_VFESPFR_RGB888;
		break;
	default:
		printk(KERN_INFO "buz: Unknown color_fmt=%x\n", btv->win.color_fmt);
		return;
	}
	if (HorDcm >= 48)
		reg |= 3 << ZR36057_VFESPFR_HFilter; /* 5 tap filter */
	else if (HorDcm >= 32)
		reg |= 2 << ZR36057_VFESPFR_HFilter; /* 4 tap filter */
	else if (HorDcm >= 16)
		reg |= 1 << ZR36057_VFESPFR_HFilter; /* 3 tap filter */
	btwrite(reg, ZR36057_VFESPFR);

	/* video display top and bottom registers */
	reg = btv->win.vidadr + btv->win.x * btv->win.bpp + btv->win.y * btv->win.bpl;
	btwrite(reg, ZR36057_VDTR);
	reg += btv->win.bpl;
	btwrite(reg, ZR36057_VDBR);

	/* display configuration */
	reg = (triton1? 0: ZR36057_VDCR_Triton)
	    | (16 << ZR36057_VDCR_MinPix)
	    | (VidWinHt << ZR36057_VDCR_VidWinHt)
	    | (VidWinWid << ZR36057_VDCR_VidWinWid);
	btwrite(reg, ZR36057_VDCR);

	/* overlay clipping mask, unused unless we later add clips */
	mask_line_size = (btv->win.width + 32) / 32;
	reg = virt_to_bus(btv->overlay_mask);
	btwrite(reg, ZR36057_MMTR);
	reg = virt_to_bus(btv->overlay_mask + mask_line_size);
	btwrite(reg, ZR36057_MMBR);
	reg = mask_line_size << ZR36057_OCR_MaskStride;
	btwrite(reg, ZR36057_OCR);

	/* video stride, status, and frame grab register */
	reg = btv->win.bpl - VidWinWid * btv->win.bpp + 3;
	if (DispMode == 0)
		reg += btv->win.bpl;
	reg &= ~3;
	printk(KERN_INFO "buz: VidWinWid=%x HStart=%x HEnd=%x HorDcm=%x DispStride=%x\n", VidWinWid, HStart, HEnd, HorDcm, reg);
	printk(KERN_INFO "buz: VidWinHt=%x VStart=%x VEnd=%x VerDcm=%x DispMode=%x\n", VidWinHt, VStart, VEnd, VerDcm, DispMode);
	reg = (reg << ZR36057_VSSFGR_DispStride);
	btwrite(reg, ZR36057_VSSFGR);
	reg |= ZR36057_VSSFGR_VidOvf;
	btwrite(reg, ZR36057_VSSFGR);
}

static void zr36057_set_jpg(struct buz *btv)
{
	struct tvnorm *tvn;
	u32 reg;
	int i;

	tvn=&tvnorms[btv->win.norm];

	/* assert P_Reset */
	btwrite(0, ZR36057_JPC);

	/* re-initialize DMA ring stuff */
	btv->grab_head = 0;
	btv->grab_tail = 0;
	for (i = 0; i < BUZ_NUM_FBUF; ++i)
	{
		btv->stat_com[i] = 1;		/* mark as unavailable to zr36057 */
	}

	/* MJPEG compression mode */
	reg	= ZR36057_JMC_JPG
		| ZR36057_JMC_CFIFO_FB
		| ZR36057_JMC_MJPGCmpMode;
	btwrite(reg, ZR36057_JMC);

	/* vertical */
	reg = (6 << ZR36057_VSP_VsyncSize) | (tvn->Ht << ZR36057_VSP_FrmTot);
	btwrite(reg, ZR36057_VSP);
	reg = (0 << ZR36057_FVAP_NAY) | (tvn->Ha << ZR36057_FVAP_PAY);
	btwrite(reg, ZR36057_FVAP);

	/* horizontal */
	reg = (tvn->Ha << ZR36057_HSP_HsyncStart) | (tvn->Wt << ZR36057_HSP_LineTot);
	btwrite(reg, ZR36057_HSP);
	reg = (10 << ZR36057_FHAP_NAX) | (tvn->Wa << ZR36057_FHAP_PAX);
	btwrite(reg, ZR36057_FHAP);

	/* field process parameters */
	reg = ZR36057_FPP_Odd_Even;
	btwrite(reg, ZR36057_FPP);

	/* code base address and FIFO threshold */
	reg = virt_to_bus(btv->stat_com);
	btwrite(reg, ZR36057_JCBA);
	reg = 0x50;
	btwrite(reg, ZR36057_JCFT);

	/* JPEG codec guest ID */
	reg = (1 << ZR36057_JCGI_JPEGuestID) | (0 << ZR36057_JCGI_JPEGuestReg);
	btwrite(reg, ZR36057_JCGI);

	/* Code transfer guest ID */
	reg = (0 << ZR36057_MCTCR_CodGuestID) | (3 << ZR36057_MCTCR_CodGuestReg);
reg |= ZR36057_MCTCR_CFlush;
	btwrite(reg, ZR36057_MCTCR);

	/* deassert P_Reset */
	btwrite(ZR36057_JPC_P_Reset, ZR36057_JPC);
}

static void zr36057_enable_jpg(struct buz *btv, int enable)
{
	/* deassert P_Reset, assert Code transfer enable */
	if (enable)
	{
		btwrite(ZR36057_ISR_GIRQ1|ZR36057_ISR_GIRQ0|ZR36057_ISR_CodRepIRQ|ZR36057_ISR_JPEGRepIRQ,ZR36057_ISR);
		zr36060_go(btv);
		btand(~ZR36057_MCTCR_CFlush, ZR36057_MCTCR);
		btor(ZR36057_JMC_Go_en, ZR36057_JMC);
		btwrite(ZR36057_JPC_P_Reset | ZR36057_JPC_CodTrnsEn | ZR36057_JPC_Active, ZR36057_JPC);
	}
	else
	{
		btor(ZR36057_MCTCR_CFlush, ZR36057_MCTCR);
		btwrite(ZR36057_JPC_P_Reset, ZR36057_JPC);
		btand(~ZR36057_JMC_Go_en, ZR36057_JMC);
	}
}

int bpp2fmt[4] = {
        VIDEO_PALETTE_HI240, VIDEO_PALETTE_RGB565, 
        VIDEO_PALETTE_RGB24, VIDEO_PALETTE_RGB32 
};

int fmt2bpp[8] = {
        0, 1, 1, 2, 3, 4, 2, 2
};

static void zr36057_set_winsize(struct buz *btv)
{
	int bpp;

	bpp = fmt2bpp[btv->win.color_fmt&0x07];
	if (btv->win.bpp == 0) 
	{
	        btv->win.bpp=bpp;
	}
	else if (btv->win.bpp!=bpp)
		btv->win.color_fmt = bpp2fmt[(btv->win.bpp-1)&3];

        zr36057_set_vfe(btv);
        zr36057_set_jpg(btv);
}

/*
The overlay mask has one bit for each pixel on a scan line, and the maximum window size
is BUZ_MAX_WIDTH*BUZ_MAX_HEIGHT pixels.
*/
static void write_overlay_mask(struct buz *btv, struct video_clip *vp, int count)
{
	unsigned mask_line_size = (btv->win.width + 31) / 32;
	u32 * mask;
	int x, y, width, height;
	unsigned i, j, k;
	u32 reg;

        printk(KERN_DEBUG "buz: overlay_mask=%p used_size=0x%x\n", btv->overlay_mask, mask_line_size * 4 * btv->win.height);

	/* fill mask with one bits */
	memset(btv->overlay_mask, ~0, mask_line_size * 4 * btv->win.height);

	for (i = 0; i < count; ++i)
	{
		/* pick up local copy of clip */
		x = vp->x;
		y = vp->y;
		width = vp->width;
		height = vp->height;

		/* trim clips that extend beyond the window */
		if (x < 0)
		{
			width += x;
			x = 0;
		}
		if (y < 0)
		{
			height += y;
			y = 0;
		}
		if (x + width > btv->win.width)
			width = btv->win.width - x;
		if (y + height > btv->win.height)
			height = btv->win.height - y;

		/* ignore degenerate clips */
		if (height <= 0)
			continue;
		if (width <= 0)
			continue;

		/* apply clip for each scan line */
		for (j = 0; j < height; ++j)
		{
			/* reset bit for each pixel */
			/* this can be optimized later if need be */
			mask = btv->overlay_mask + (y + j) * mask_line_size;
			for (k = 0; k < width; ++k)
			{
				mask[(x + k) / 32] &= ~((u32)1 << (x + k) % 32);
			}
		}
	}

	/* enable overlay */
	reg = (mask_line_size << ZR36057_OCR_MaskStride) | ZR36057_OCR_OvlEnable;
	btwrite(reg, ZR36057_OCR);
}



/*
 *	Grab into virtual memory.
 *	Currently only does double buffering. Do we need more?
 */

static int vgrab(struct buz *btv, struct video_mmap *mp)
{
	unsigned frame = mp->frame;
	unsigned i;

	/*
	 *	No grabbing past the end of the buffer!
	 */

	if (frame >= BUZ_NUM_FBUF || frame < 0)
		return -EINVAL;

	if (mp->height < BUZ_MIN_HEIGHT || mp->width < BUZ_MIN_WIDTH)
		return -EINVAL;

	if (mp->height > BUZ_MAX_HEIGHT || mp->width > BUZ_MAX_WIDTH)
		return -EINVAL;

	/* FIXME: for now format just indicates maximum frame size, or zero to continue grabbing with same parameters */
	if (mp->format > BUZ_MAX_FBUF/1024)
		return -EINVAL;

	if(btv->fbuffer==NULL)
	{
		//if(fbuffer_alloc(btv))
			return -ENOBUFS;
	}

	/* make sure a grab isn't going on currently with this buffer */
	if (btv->frag_tab_busy[frame])
		return -EAGAIN;

	/* FIXME: for now format just indicates maximum frame size, or zero to continue grabbing with same parameters */
	if (mp->format != 0)
	{
		struct tvnorm *tvn;

		/* new sequence, set up JPEG stuff */
		tvn = &tvnorms[btv->win.norm];

		/*
		 *	FIXME: Check the format of the grab here. This is probably
		 *	also less restrictive than the normal overlay grabs. Stuff
		 *	like QCIF has meaning as a capture.
		 */

		/* actually it is *more* restrictive.  Full, 1/2, or 1/4 size only. */
		if (!((mp->width == tvn->Wa && mp->height == tvn->Ha)
		||   (mp->width == tvn->Wa/2 && mp->height == tvn->Ha/2)
		||   (mp->width == tvn->Wa/4 && mp->height == tvn->Ha/4)))
			return -EINVAL;

		btv->gwidth = mp->width;
		btv->gheight = mp->height;
		btv->gfmt = mp->format;

		/*
		 *	Ok load up the zr36060
		 */
		zr36057_enable_jpg(btv, 0);
		zr36060_set_cap(btv);
		zr36057_set_jpg(btv);
	}

	/* since there was at least one unused buffer there's room for at least one more stat_com entry */
	i = btv->grab_tail++ & BUZ_MASK_FBUF;
	btv->stat_com[i] = btv->frag_tab_bus[frame];
	btv->stat_com_frame[i] = frame;
	btv->stat_com_result[i] = -EINTR;
	btv->stat_com_busy[i] = 1;
	btv->frag_tab_index[frame] = i;
	btv->frag_tab_busy[frame] = 1;
	printk(KERN_INFO "buz: stat_com[%d]=%x stat_com_frame[%d]=%d frag_tab_index[%d]=%d\n",
			i, btv->stat_com[i],
			i, btv->stat_com_frame[i],
			frame, btv->frag_tab_index[frame]);

	if (mp->format != 0)
		zr36057_enable_jpg(btv, 1);

	return 0;
}

static void buz_reap_stat_com(struct buz * btv)
{
	while (btv->grab_head != btv->grab_tail)
	{
		int i = btv->grab_head & BUZ_MASK_FBUF;

		IDEBUG(printk ("buz: i=%d grab_head=%x grab_tail=%x stat_com[]=%08x %08x %08x %08x\n",
					i, btv->grab_head, btv->grab_tail,
					btv->stat_com[0], btv->stat_com[1], btv->stat_com[2], btv->stat_com[3]));
		IDEBUG(printk ("buz: stat_com[i]=%08x stat_com_frame[i]=%x\n", btv->stat_com[i], btv->stat_com_frame[i]));
		if ((btv->stat_com[i] & 1) == 0)
			return;
		btv->stat_com_result[i] = (btv->stat_com[i] & 0x7fffff) >> 1;
		btv->stat_com_busy[i] = 0;
		wake_up_interruptible(&btv->stat_com_capq[i]);
		++btv->grab_head;
	}
}

static int vsync(struct buz * btv, int frame)
{
	unsigned i;
	int res;

	if (frame >= BUZ_NUM_FBUF || frame < 0)
		return -EINVAL;

	if (!btv->frag_tab_busy[frame])
		return -EINVAL;

	i = btv->frag_tab_index[frame];
	printk(KERN_INFO "buz: stat_com[%d]=%x stat_com_frame[%d]=%d frag_tab_index[%d]=%d\n",
			i, btv->stat_com[i],
			i, btv->stat_com_frame[i],
			frame, btv->frag_tab_index[frame]);
	if (btv->stat_com_busy[i])
		interruptible_sleep_on(&btv->stat_com_capq[i]);
	printk(KERN_INFO "buz: stat_com[%d]=%x stat_com_frame[%d]=%d frag_tab_index[%d]=%d\n",
			i, btv->stat_com[i],
			i, btv->stat_com_frame[i],
			frame, btv->frag_tab_index[frame]);

	buz_reap_stat_com(btv);
	res = btv->stat_com_result[i];
	btv->frag_tab_busy[frame] = 0;
	return res;
}

struct vidbases 
{
	unsigned short vendor, device;
	char *name;
	uint badr;
};

static struct vidbases vbs[] = {
	{ PCI_VENDOR_ID_TSENG, 0, "TSENG", PCI_BASE_ADDRESS_0},
	{ PCI_VENDOR_ID_MATROX, PCI_DEVICE_ID_MATROX_MIL,
		"Matrox Millennium", PCI_BASE_ADDRESS_1},
	{ PCI_VENDOR_ID_MATROX, 0x051a, "Matrox Mystique", PCI_BASE_ADDRESS_1},
	{ PCI_VENDOR_ID_S3, 0, "S3", PCI_BASE_ADDRESS_0},
	{ PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_ATI_210888GX,
		"ATI MACH64 Winturbo", PCI_BASE_ADDRESS_0},
	{ PCI_VENDOR_ID_ATI, PCI_DEVICE_ID_ATI_215GTB,
		"ATI MACH64 RageII+", PCI_BASE_ADDRESS_0},
	{ PCI_VENDOR_ID_CIRRUS, 0, "Cirrus Logic", PCI_BASE_ADDRESS_0},
	{ PCI_VENDOR_ID_N9, PCI_DEVICE_ID_N9_I128, 
		"Number Nine Imagine 128", PCI_BASE_ADDRESS_0},
	{ PCI_VENDOR_ID_DEC, PCI_DEVICE_ID_DEC_TGA,
		"DEC DC21030", PCI_BASE_ADDRESS_0},
};


/* DEC TGA offsets stolen from XFree-3.2 */

static uint dec_offsets[4] = {
	0x200000,
	0x804000,
	0,
	0x1004000
};

#define NR_CARDS (sizeof(vbs)/sizeof(struct vidbases))

/* Scan for PCI display adapter
   if more than one card is present the last one is used for now */

static int find_vga(void)
{
	unsigned int class;
	unsigned short badr;
	int found=0, i, tga_type;
	unsigned int vidadr=0;
	struct pci_dev *dev;


	for (dev = pci_devices; dev != NULL; dev = dev->next)
	{
		class = dev->class >> 8;
/*		if (class == PCI_CLASS_DISPLAY_VGA) {*/
		if ((class>>8) == PCI_BASE_CLASS_DISPLAY ||
			/* Number 9 GXE64Pro needs this */
			class == PCI_CLASS_NOT_DEFINED_VGA) 
		{
			badr=0;
			printk(KERN_INFO "buz: PCI display adapter: ");
			for (i=0; i<NR_CARDS; i++) 
			{
				if (dev->vendor==vbs[i].vendor) 
				{
					if (vbs[i].device) 
						if (vbs[i].device!=dev->device)
							continue;
					printk("%s.\n", vbs[i].name);
					badr=vbs[i].badr;
					break;
				}
			}
			if (!badr) 
			{
				printk(KERN_ERR "buz: Unknown video memory base address.\n");
				continue;
			}
			pci_read_config_dword(dev, badr, &vidadr);
			if (vidadr & PCI_BASE_ADDRESS_SPACE_IO) 
			{
				printk(KERN_ERR "buz: Memory seems to be I/O memory.\n");
				printk(KERN_ERR "buz: Check entry for your card type in buz.c vidbases struct.\n");
				continue;
			}
			vidadr &= PCI_BASE_ADDRESS_MEM_MASK;
			if (!vidadr) 
			{
				printk(KERN_ERR "buz: Memory @ 0, must be something wrong!");
				continue;
			}

			if (dev->vendor==PCI_VENDOR_ID_DEC)
				if (dev->device==PCI_DEVICE_ID_DEC_TGA) 
			{
				tga_type = (readl((unsigned long)vidadr) >> 12) & 0x0f;
				if (tga_type != 0 && tga_type != 1 && tga_type != 3) 
				{
					printk(KERN_ERR "buz: TGA type (0x%x) unrecognized!\n", tga_type);
					found--;
				}
				vidadr+=dec_offsets[tga_type];
			}

			DEBUG(printk(KERN_DEBUG "buz: memory @ 0x%08x, ", vidadr));
			DEBUG(printk(KERN_DEBUG "devfn: 0x%04x.\n", dev->devfn));
			found++;
		}
	}

	if (vidmem)
	{
		vidadr=vidmem<<20;
		printk(KERN_INFO "buz: Video memory override: 0x%08x\n", vidadr);
		found=1;
	}
	for (i=0; i<BUZ_MAX; i++)
		buzs[i].win.vidadr=vidadr;

	return found;
}

#define  TRITON_PCON	           0x50 
#define  TRITON_BUS_CONCURRENCY   (1<<0)
#define  TRITON_STREAMING	  (1<<1)
#define  TRITON_WRITE_BURST	  (1<<2)
#define  TRITON_PEER_CONCURRENCY  (1<<3)

static void handle_chipset(void)
{
	int index;

	for (index = 0; index < 8; index++)
	{
		unsigned char bus, devfn;
		unsigned char b;

		/* Beware the SiS 85C496 my friend - rev 49 don't work with a buz */

		if (!pcibios_find_device(PCI_VENDOR_ID_SI, 
					 PCI_DEVICE_ID_SI_496, 
					 index, &bus, &devfn))
		{
			printk(KERN_WARNING "Buz and SIS 85C496 chipset don't always work together.\n");
		}			

		if (!pcibios_find_device(PCI_VENDOR_ID_INTEL, 
					 PCI_DEVICE_ID_INTEL_82441,
					 index, &bus, &devfn)) 
		{
			pcibios_read_config_byte(bus, devfn, 0x53, &b);
			DEBUG(printk(KERN_INFO "buz: Host bridge: 82441FX Natoma, "));
			DEBUG(printk("bufcon=0x%02x\n",b));
		}

		if (!pcibios_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82437,
			    index, &bus, &devfn)) 
		{
			printk(KERN_INFO "buz: Host bridge 82437FX Triton PIIX\n");
			triton1=1;

#if 0			
			/* The ETBF bit SHOULD make all this unnecessary */
			/* 430FX (Triton I) freezes with bus concurrency on -> switch it off */
			{   
			        unsigned char bo;

				pcibios_read_config_byte(bus, devfn, TRITON_PCON, &b);
				bo=b;
				DEBUG(printk(KERN_DEBUG "buz: 82437FX: PCON: 0x%x\n",b));

				if(!(b & TRITON_BUS_CONCURRENCY)) 
				{
					printk(KERN_WARNING "buz: 82437FX: disabling bus concurrency\n");
					b |= TRITON_BUS_CONCURRENCY;
				}

				if(b & TRITON_PEER_CONCURRENCY) 
				{
					printk(KERN_WARNING "buz: 82437FX: disabling peer concurrency\n");
					b &= ~TRITON_PEER_CONCURRENCY;
				}
				if(!(b & TRITON_STREAMING)) 
				{
					printk(KERN_WARNING "buz: 82437FX: disabling streaming\n");
					b |=  TRITON_STREAMING;
				}

				if (b!=bo) 
				{
					pcibios_write_config_byte(bus, devfn, TRITON_PCON, b); 
					printk(KERN_DEBUG "buz: 82437FX: PCON changed to: 0x%x\n",b);
				}
			}
#endif
		}
	}
}

static void buz_irq(int irq, void *dev_id, struct pt_regs * regs)
{
	u32 stat,astat;
	u32 dstat;
	int count;
	struct buz *btv;

	btv=(struct buz *)dev_id;
	count=0;
	while (1) 
	{
		/* get/clear interrupt status bits */
		stat = btread(ZR36057_ISR);
		astat = stat & (ZR36057_ISR_GIRQ1|ZR36057_ISR_GIRQ0|ZR36057_ISR_CodRepIRQ|ZR36057_ISR_JPEGRepIRQ);
		if (!astat)
			return;
		btwrite(astat,ZR36057_ISR);
		IDEBUG(printk ("buz: astat %08x\n",astat));
		IDEBUG(printk ("buz:  stat %08x\n",stat));

		/* get device status bits */
		dstat=btread(ZR36057_ISR);

		IDEBUG(printk ("buz: interrupt 57_VSSFGR=%08x 57_MCTCR=%08x 60_code=%02x 60_intr=%02x\n",
			btread(ZR36057_VSSFGR), btread(ZR36057_MCTCR), zr36060_read_8(btv, 0x001), zr36060_read_8(btv, 0x008)));

		if (astat&ZR36057_ISR_GIRQ0) 
		{
			IDEBUG(printk ("buz: ZR36057_ISR_GIRQ0\n"));
			btand(~ZR36057_ICR_GIRQ0, ZR36057_ICR);
		}
		if (astat&ZR36057_ISR_GIRQ1) 
		{
			IDEBUG(printk ("buz: ZR36057_ISR_GIRQ1\n"));
			btand(~ZR36057_ICR_GIRQ1, ZR36057_ICR);
		}
		if (astat&ZR36057_ISR_CodRepIRQ) 
		{
			IDEBUG(printk ("buz: ZR36057_ISR_CodRepIRQ\n"));
			btand(~ZR36057_ICR_CodRepIRQ, ZR36057_ICR);
		}
		if (astat&ZR36057_ISR_JPEGRepIRQ) 
		{
			IDEBUG(printk ("buz: ZR36057_ISR_JPEGRepIRQ\n"));
			IDEBUG(printk ("buz: SF=%x\n", zr36060_read_16(btv, 0x011)));
			buz_reap_stat_com(btv);
		}

		count++;
		if (count > 10)
			printk (KERN_WARNING "buz: irq loop %d\n", count);
		if (count > 20) 
		{
			btwrite(0,ZR36057_ICR);
			printk(KERN_ERR 
			       "buz: IRQ lockup, cleared int mask\n");
		}
	}
}



/*
 *	Open a buz card. Right now the flags stuff is just playing
 */

static int buz_open(struct video_device *dev, int flags)
{
	struct buz *btv = (struct buz *)dev;
	int users, i;

	switch (flags)
	{
		case 0:
			if (btv->user)
				return -EBUSY;
			btv->user++;
			btv->got_buffer = 0;
			btv->got_window = 0;

			for (i=users=0; i<buz_num; i++)
				users+=buzs[i].user;
			if (users==1)
				find_vga();
			if (fbuffer_alloc(btv) < 0)
			{
				btv->user--;
				return -ENOMEM;
			}

			zr36057_enable_jpg(btv, 0);

			btwrite(ZR36057_ISR_GIRQ1|ZR36057_ISR_GIRQ0|ZR36057_ISR_CodRepIRQ|ZR36057_ISR_JPEGRepIRQ, ZR36057_ISR);
			btor(ZR36057_ICR_IntPinEn, ZR36057_ICR);
			break;
		case 1:
			break;
	}
	MOD_INC_USE_COUNT;
	return 0;   
}

static void buz_close(struct video_device *dev)
{
	struct buz *btv=(struct buz *)dev;
	int i;

	/* disable interrupts */
	btand(~ZR36057_ICR_IntPinEn, ZR36057_ICR);

	/* wake up sleeping beauties */
	for (i = 0; i < BUZ_NUM_FBUF; ++i)
	{
		wake_up_interruptible(&btv->stat_com_capq[i]);
	}

	zr36057_enable_jpg(btv, 0);
	zr36057_cap(btv, 0);
	btv->user--;

	if (btv->fbuffer)
	{
		rvfree((void *) btv->fbuffer, BUZ_NUM_FBUF*BUZ_MAX_FBUF);
		btv->fbuffer = 0;
	}

	MOD_DEC_USE_COUNT;  
}


static long buz_write(struct video_device *v, const char *buf, unsigned long count, int nonblock)
{
	return -EINVAL;
}

static long buz_read(struct video_device *v, char *buf, unsigned long count, int nonblock)
{
	return -EINVAL;
}

/*
 *	Helper for adding clips.
 */

static void new_risc_clip(struct video_window *vw, struct video_clip *vcp, int x, int y, int w, int h)
{
	vcp[vw->clipcount].x=x;
	vcp[vw->clipcount].y=y;
	vcp[vw->clipcount].width=w;
	vcp[vw->clipcount].height=h;
	vw->clipcount++;
}

/*
 *	ioctl routine
 */


static int buz_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct buz *btv=(struct buz *)dev;

	switch (cmd)
	{	
		case VIDIOCGCAP:
		{
			struct video_capability b;
			strcpy(b.name,btv->video_dev.name);
			b.type = VID_TYPE_CAPTURE|
				VID_TYPE_OVERLAY|
				VID_TYPE_CLIPPING|
				VID_TYPE_FRAMERAM|
				VID_TYPE_SCALES;
			b.channels = 2; /* composite, svhs */
			b.audios = 0;
			b.maxwidth = BUZ_MAX_WIDTH;
			b.maxheight = BUZ_MAX_HEIGHT;
			b.minwidth = BUZ_MIN_WIDTH;
			b.minheight = BUZ_MIN_HEIGHT;
			if(copy_to_user(arg,&b,sizeof(b)))
				return -EFAULT;
			return 0;
		}
		case VIDIOCGCHAN:
		{
			struct video_channel v;
			if(copy_from_user(&v, arg,sizeof(v)))
				return -EFAULT;
			v.flags=0;
			v.tuners=0;
			v.type=VIDEO_TYPE_CAMERA;
			switch(v.channel)
			{
				case 0:
					strcpy(v.name,"Composite");
					break;
				case 1:
					strcpy(v.name,"SVHS");
					break;
				default:
					return -EINVAL;
			}
			if(copy_to_user(arg,&v,sizeof(v)))
				return -EFAULT;
			return 0;
		}
		case VIDIOCSCHAN:
		{
			int v;
			if(copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			switch(v)
			{
				case 0:
					v = 3;
					break;
				case 1:
					v = 5;
					break;
				default:
					return -EINVAL;
			}
			i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_INPUT, &v);
			return 0;
		}
		case VIDIOCGTUNER:
		{
			return -EINVAL;
		}
		case VIDIOCSTUNER:
		{
			return -EINVAL;
		}
		case VIDIOCGPICT:
		{
			struct video_picture p=btv->picture;
			if(btv->win.bpp==2)
				p.palette=VIDEO_PALETTE_RGB565;
			if(btv->win.bpp==3)
				p.palette=VIDEO_PALETTE_RGB24;
			if(btv->win.bpp==4)
				p.palette=VIDEO_PALETTE_RGB32;

			if(copy_to_user(arg, &p, sizeof(p)))
				return -EFAULT;
			return 0;
		}
		case VIDIOCSPICT:
		{
			struct video_picture p;
			int v;
			if(copy_from_user(&p, arg,sizeof(p)))
				return -EFAULT;
			v = p.brightness;
			i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_BRIGHTNESS, &v);
			v = p.contrast;
			i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_CONTRAST, &v);
			v = p.hue;
			i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_HUE, &v);
			v = p.colour;
			i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_SATURATION, &v);
			btv->picture=p;
			return 0;
		}
		case VIDIOCSWIN:
		{
			struct video_window vw;
			int on;

			if(copy_from_user(&vw,arg,sizeof(vw)))
				return -EFAULT;

			if(vw.flags)
				return -EINVAL;

			btv->win.x=vw.x;
			btv->win.y=vw.y;
			btv->win.width=vw.width;
			btv->win.height=vw.height;

			if(btv->win.height>btv->win.cropheight/2)
				btv->win.interlace=1;
			else
				btv->win.interlace=0;

			on=btv->cap;

			zr36057_cap(btv,0);
			zr36057_set_winsize(btv);

			if(vw.clipcount>256)
				return -EDOM;	/* Too many! */

			/*
			 *	Do any clips.
			 */

			if (vw.clipcount)
			{
				struct video_clip * vcp = vmalloc(sizeof(struct video_clip)*(vw.clipcount+4));
				if(vcp==NULL)
					return -ENOMEM;
				if(copy_from_user(vcp,vw.clips,sizeof(struct video_clip)*vw.clipcount))
				{
					vfree(vcp);
					return -EFAULT;
				}

				/*
				 *	Impose display clips
				 */
				if(btv->win.x<0)
					new_risc_clip(&vw, vcp, 0, 0, -btv->win.x, btv->win.height-1);
				if(btv->win.y<0)
					new_risc_clip(&vw, vcp, 0, 0, btv->win.width-1,-btv->win.y);
				if(btv->win.x+btv->win.width> btv->win.swidth)
					new_risc_clip(&vw, vcp, btv->win.swidth-btv->win.x, 0, btv->win.width-1, btv->win.height-1);
				if(btv->win.y+btv->win.height > btv->win.sheight)
					new_risc_clip(&vw, vcp, 0, btv->win.sheight-btv->win.y, btv->win.width-1, btv->win.height-1);

				write_overlay_mask(btv,vcp, vw.clipcount);
				vfree(vcp);
			}

			btv->got_window = 1;

			if(on && btv->win.vidadr!=0)
				zr36057_cap(btv,1);
			return 0;
		}
		case VIDIOCGWIN:
		{
			struct video_window vw;
			/* Oh for a COBOL move corresponding .. */
			vw.x=btv->win.x;
			vw.y=btv->win.y;
			vw.width=btv->win.width;
			vw.height=btv->win.height;
			vw.chromakey=0;
			vw.flags=0;
			if(btv->win.interlace)
				vw.flags|=VIDEO_WINDOW_INTERLACE;
			if(copy_to_user(arg,&vw,sizeof(vw)))
				return -EFAULT;
			return 0;
		}
		case VIDIOCCAPTURE:
		{
			int v;
			if(copy_from_user(&v, arg,sizeof(v)))
				return -EFAULT;
			if(v==0)
			{
				zr36057_cap(btv,0);
			}
			else
			{
			        if(btv->win.vidadr==0 || btv->win.width==0
				   || btv->win.height==0)
				  return -EINVAL;
				zr36057_cap(btv,1);
			}
			return 0;
		}
		case VIDIOCGFBUF:
		{
			struct video_buffer v;
			v.base=(void *)btv->win.vidadr;
			v.height=btv->win.sheight;
			v.width=btv->win.swidth;
			v.depth=btv->win.depth;
			v.bytesperline=btv->win.bpl;
			if(copy_to_user(arg, &v,sizeof(v)))
				return -EFAULT;
			return 0;

		}
		case VIDIOCSFBUF:
		{
			struct video_buffer v;
			if(!suser())
				return -EPERM;
			if(copy_from_user(&v, arg,sizeof(v)))
				return -EFAULT;
			if(v.depth!=16 && v.depth!=24 && v.depth!=32)
				return -EINVAL;
			if (v.base)
			{
			        /* also handle virtual base addresses */
			        if ((unsigned int)v.base>=0xe0000000UL)
			                btv->win.vidadr = (uint)v.base;
				else 
					btv->win.vidadr = (uint)__va(uvirt_to_bus((uint)v.base));
			}
			btv->win.sheight=v.height;
			btv->win.swidth=v.width;
			btv->win.bpp=((v.depth+7)&0x38)/8;
			btv->win.depth=v.depth;
			btv->win.bpl=v.bytesperline;

			DEBUG(printk("Display at %p is %d by %d, bytedepth %d, bpl %d\n",
				v.base, v.width,v.height, btv->win.bpp, btv->win.bpl));
			//zr36057_set_winsize(btv);
			btv->got_buffer = 1;
			return 0;		
		}
		case VIDIOCKEY:
		{
			/* Will be handled higher up .. */
			return 0;
		}
		case VIDIOCGFREQ:
		{
			return -EINVAL;
		}
		case VIDIOCSFREQ:
		{
			return -EINVAL;
		}

		case VIDIOCGAUDIO:
		{
			return -EINVAL;
		}
		case VIDIOCSAUDIO:
		{
			return -EINVAL;
		}

	        case VIDIOCSYNC:
		{
			int v;
			if(copy_from_user(&v, arg,sizeof(v)))
				return -EFAULT;
		        return vsync(btv, v);
		}

	        case VIDIOCMCAPTURE:
		{
                        struct video_mmap vm;
			if(copy_from_user((void *) &vm, (void *) arg, sizeof(vm)))
				return -EFAULT;
		        return vgrab(btv, &vm);
		}
		default:
			return -ENOIOCTLCMD;
	}
	return 0;
}


/*
 *	This maps the vmalloced and reserved fbuffer to user space.
 *
 *  FIXME: 
 *  - PAGE_READONLY should suffice!?
 *  - remap_page_range is kind of inefficient for page by page remapping.
 *    But e.g. pte_alloc() does not work in modules ... :-(
 */

static int buz_mmap(struct video_device *dev, const char *adr, unsigned long size)
{
	struct buz *btv=(struct buz *)dev;
        unsigned long start=(unsigned long) adr;
	unsigned long page,pos;

	if (size>BUZ_NUM_FBUF*BUZ_MAX_FBUF)
	        return -EINVAL;
	if (!btv->fbuffer)
	{
		//if(fbuffer_alloc(btv))
			return -ENOBUFS;
	}
	pos=(unsigned long) btv->fbuffer;
	while (size > 0) 
	{
	        page = kvirt_to_phys(pos);
		if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED))
		        return -EAGAIN;
		start+=PAGE_SIZE;
		pos+=PAGE_SIZE;
		size-=PAGE_SIZE;    
	}
	return 0;
}

static int buz_init_done(struct video_device *dev)
{
	return 0;
}

static struct video_device buz_template=
{
	"UNSET",
	VID_TYPE_CAPTURE|VID_TYPE_OVERLAY,
	VID_HARDWARE_BT848,
	buz_open,
	buz_close,
	buz_read,
	buz_write,
	buz_ioctl,
	buz_mmap,
	buz_init_done,
	NULL,
	0,
	0
};

/*
Wait til post office is no longer busy */
static int post_office_wait(struct buz *btv)
{
	u32 por;

	while (((por = btread(ZR36057_POR)) & (ZR36057_POR_POPen | ZR36057_POR_POTime)) == ZR36057_POR_POPen)
	{
		/* wait for something to happen */
	}
	if ((por & ZR36057_POR_POPen) != 0)
	{
		printk(KERN_INFO "buzpop: pending %08x\n", por);
		return -1;
	}
	if ((por & (ZR36057_POR_POTime | ZR36057_POR_POPen)) != 0)
	{
		printk(KERN_INFO "buzpop: timeout %08x\n", por);
		return -1;
	}
	return 0;
}

static int post_office_write(struct buz *btv, unsigned guest, unsigned reg, unsigned value)
{
	u32 por;

	por = ZR36057_POR_PODir | ZR36057_POR_POTime | ((guest & 7) << 20) | ((reg & 7) << 16) | (value & 0xFF);
	btwrite(por, ZR36057_POR);
	return post_office_wait(btv);
}

static int post_office_read(struct buz *btv, unsigned guest, unsigned reg)
{
	u32 por;

	por = ZR36057_POR_POTime | ((guest & 7) << 20) | ((reg & 7) << 16);
	btwrite(por, ZR36057_POR);
	if (post_office_wait(btv) < 0)
	{
		return -1;
	}
	return btread(ZR36057_POR) & 0xFF;
}

static int zr36060_write_8(struct buz *btv, unsigned reg, unsigned val)
{
	if (post_office_wait(btv)
	||  post_office_write(btv, 0, 1, reg >> 8)
	||  post_office_write(btv, 0, 2, reg))
		return -1;
	return post_office_write(btv, 0, 3, val);
}

static int zr36060_write_16(struct buz *btv, unsigned reg, unsigned val)
{
	if (zr36060_write_8(btv, reg + 0, val >> 8))
		return -1;
	return zr36060_write_8(btv, reg + 1, val >> 0);
}

static int zr36060_write_24(struct buz *btv, unsigned reg, unsigned val)
{
	if (zr36060_write_8(btv, reg + 0, val >> 16))
		return -1;
	return zr36060_write_16(btv, reg + 1, val >> 0);
}

static int zr36060_write_32(struct buz *btv, unsigned reg, unsigned val)
{
	if (zr36060_write_16(btv, reg + 0, val >> 16))
		return -1;
	return zr36060_write_16(btv, reg + 2, val >> 0);
}

static u32 zr36060_read_8(struct buz *btv, unsigned reg)
{
	if (post_office_wait(btv)
	||  post_office_write(btv, 0, 1, reg >> 8)
	||  post_office_write(btv, 0, 2, reg))
		return -1;
	return post_office_read(btv, 0, 3) & 0xFF;
}

static u32 zr36060_read_16(struct buz *btv, unsigned reg)
{
	return zr36060_read_8(btv, reg + 1) | zr36060_read_8(btv, reg + 0) * 256;
}

static u32 zr36060_read_24(struct buz *btv, unsigned reg)
{
	return zr36060_read_16(btv, reg + 1) | zr36060_read_8(btv, reg + 0) * 256 * 256;
}

static u32 zr36060_read_32(struct buz *btv, unsigned reg)
{
	return zr36060_read_16(btv, reg + 2) | zr36060_read_16(btv, reg + 0) * 256 * 256;
}

static int zr36060_start(struct buz *btv)
{
	return post_office_write(btv, 1, 0, 0);
}

static int zr36060_reset(struct buz *btv)
{
	return post_office_write(btv, 3, 0, 0);
}

static void zr36060_sleep(struct buz *btv, int sleep)
{
	GPIO(btv, 1, !sleep);
}

static int zr36060_go(struct buz *btv)
{
	return post_office_write(btv, 1, 0, 0);
}

static void zr36060_dump(struct buz * btv)
{
	int i, j;
	unsigned char line[16];
	for (i = 0; i < 1024; i += 16) {
		printk(KERN_INFO "060: %03x", i);
		for (j = 0; j < 16; ++j) {
			line[j] = zr36060_read_8(btv, i + j);
			printk(" %02x", line[j]);
		}
		printk(" |");
		for (j = 0; j < 16; ++j) {
			printk("%c", (line[j] > ' ' && line[j] < 127)? line[j]: '.');
		}
		printk("|\n");
	}
}

static void zr36060_set_jpg(struct buz * btv)
{
	struct tvnorm *tvn;
	u32 reg;

	tvn = &tvnorms[btv->win.norm];

	reg	= (1 << 0)	/* CodeMstr */
		| (0 << 2)	/* CFIS=0 */
		| (0 << 6)	/* Endian=0 */
		| (0 << 7);	/* Code16=0 */
	zr36060_write_8(btv, 0x002, reg);

	reg	= 0xa4;		/* Codec mode = compression with variable scale factor */
	zr36060_write_8(btv, 0x003, reg);

	reg	= 0x00;		/* reerved, mbz */
	zr36060_write_8(btv, 0x004, reg);

	reg	= 0xf0;		/* 480 bits/block */
	zr36060_write_8(btv, 0x005, reg);

	reg	= (1 << 7)	/* APP=1 */
		| (1 << 6)	/* COM=1 */
		| (0 << 5)	/* DRI=0 */
		| (1 << 4)	/* DQT=1 */
		| (1 << 3);	/* DHT=1 */
	zr36060_write_8(btv, 0x006, reg);

	reg	= (0 << 3)	/* DATERR=0 */
		| (0 << 2)	/* END=0 */
		| (0 << 1)	/* EOI=0 */
		| (0 << 0);	/* EOAV=0 */
	zr36060_write_8(btv, 0x007, reg);

	reg	= (btv->gfmt - 1) * 1024; /* TCV_NET */
	zr36060_write_32(btv, 0x009, reg);

	reg	= 0x001d;	/* SF=0.1d */
	reg	= 0x0100;	/* SF=1.0 */
	zr36060_write_16(btv, 0x011, reg);

	reg	= 0x00ffffff;	/* AF=max */
	zr36060_write_24(btv, 0x013, reg);

	reg	= 0x0000;	/* test */
	zr36060_write_16(btv, 0x024, reg);
}

static void zr36060_set_video(struct buz * btv)
{
	struct tvnorm *tvn;
	u32 reg;

	tvn = &tvnorms[btv->win.norm];

	reg	= (0 << 7)	/* Video8=0 */
		| (1 << 6)	/* Range=1 */
		| (0 << 3)	/* FlDet=0 */
		| (1 << 2)	/* FlVedge=1 */
		| (1 << 1)	/* FlExt=1 */
		| (0 << 0);	/* SyncMstr=0 */
	zr36060_write_8(btv, 0x030, reg);

	reg	= (1 << 7)	/* VCLKPol=0 */
		| (0 << 6)	/* PValPol=0 */
		| (0 << 5)	/* PoePol=0 */
		| (0 << 4)	/* SImgPol=0 */
		| (0 << 3)	/* BLPol=0 */
		| (0 << 2)	/* FlPol=0 */
		| (0 << 1)	/* HSPol=0 */
		| (1 << 0);	/* VSPol=1 */
	zr36060_write_8(btv, 0x031, reg);

	if (btv->gwidth == tvn->Wa)
		reg = (0 << 0);	/* HScale=0 */
	else if (btv->gwidth == tvn->Wa/2)
		reg = (1 << 0);	/* HScale=0 */
	else /* if (btv->gwidth == tvn->Wa/4) */
		reg = (2 << 0);	/* HScale=0 */
	if (btv->gheight != tvn->Ha)
		reg |= (1 << 2);/* VScale=1 */
	zr36060_write_8(btv, 0x032, reg);

	reg	= 0x0a;		/* BackY=10 */
	zr36060_write_8(btv, 0x033, reg);

	reg	= 0x0a;		/* BackU=10 */
	zr36060_write_8(btv, 0x034, reg);

	reg	= 0x0a;		/* BackV=10 */
	zr36060_write_8(btv, 0x035, reg);

#if 0 /* not used yet */
	/* sync generator */
	reg	= tvn->Ht - 1;	/* Vtotal */
	zr36060_write_16(btv, 0x036, reg);

	reg	= tvn->Wt - 1;	/* Htotal */
	zr36060_write_16(btv, 0x038, reg);
#endif

	/* active area */
	reg	= 0x00c;	/* Vstart=12 */
	zr36060_write_16(btv, 0x042, reg);

	reg	= 0x00c + tvn->Ha/2; /* Vend */
	zr36060_write_16(btv, 0x044, reg);

	reg	= 0x06b;	/* Hstart=107 */
	zr36060_write_16(btv, 0x046, reg);

	reg	= 0x06b + tvn->Wa; /* Hend */
	zr36060_write_16(btv, 0x048, reg);

	/* subimage area */
	reg	= 0x00c;	/* SVstart=12 */
	zr36060_write_16(btv, 0x04a, reg);

	reg	= 0x00c + tvn->Ha/2; /* SVend */
	zr36060_write_16(btv, 0x04c, reg);

	reg	= 0x06b;	/* SHstart=107 */
	zr36060_write_16(btv, 0x04e, reg);

	reg	= 0x06b + tvn->Wa; /* SHend */
	zr36060_write_16(btv, 0x050, reg);
}

static void zr36060_set_jpg_SOF(struct buz * btv)
{
	struct tvnorm *tvn;
	u32 reg;

	tvn = &tvnorms[btv->win.norm];

	reg	= 0xffc0;	/* SOF marker */
	zr36060_write_16(btv, 0x060, reg);

	reg	= 17;		/* SOF length */
	zr36060_write_16(btv, 0x062, reg);

	reg	= 8;		/* precision 8 bits */
	zr36060_write_8(btv, 0x064, reg);

	reg	= tvn->Ha;	/* image height */
	zr36060_write_16(btv, 0x065, reg);

	reg	= tvn->Wa;	/* image width */
	zr36060_write_16(btv, 0x067, reg);

	reg	= 3;		/* 3 color components */
	zr36060_write_8(btv, 0x069, reg);

	reg	= 0x002100;	/* Y component */
	zr36060_write_24(btv, 0x06a, reg);

	reg	= 0x011101;	/* U component */
	zr36060_write_24(btv, 0x06d, reg);

	reg	= 0x021101;	/* V component */
	zr36060_write_24(btv, 0x070, reg);
}

static void zr36060_set_jpg_SOS(struct buz * btv)
{
	struct tvnorm *tvn;
	u32 reg;

	tvn = &tvnorms[btv->win.norm];

	reg	= 0xffda;	/* SOS marker */
	zr36060_write_16(btv, 0x07a, reg);

	reg	= 12;		/* SOS length */
	zr36060_write_16(btv, 0x07c, reg);

	reg	= 3;		/* 3 color components */
	zr36060_write_8(btv, 0x07e, reg);

	reg	= 0x0000;	/* Y component */
	zr36060_write_16(btv, 0x07f, reg);

	reg	= 0x0111;	/* U component */
	zr36060_write_16(btv, 0x081, reg);

	reg	= 0x0211;	/* V component */
	zr36060_write_16(btv, 0x083, reg);

	reg	= 0x003f00;	/* Start, end spectral scans */
	zr36060_write_24(btv, 0x085, reg);
}

static void zr36060_set_jpg_DRI(struct buz * btv)
{
	struct tvnorm *tvn;
	u32 reg;

	tvn = &tvnorms[btv->win.norm];

	reg	= 0xffdd;	/* DRI marker */
	zr36060_write_16(btv, 0x0c0, reg);

	reg	= 4;		/* DRI length */
	zr36060_write_16(btv, 0x0c2, reg);

	reg	= 8;		/* length in MCUs */
	zr36060_write_16(btv, 0x0c4, reg);
}

static void zr36060_set_jpg_DQT(struct buz * btv)
{
	unsigned i;
	unsigned adr;
	static const u8 dqt[] = {
		0xff, 0xdb,	/* DHT marker */
		0x00, 0x84,	/* DHT length */
		0x00,		/* table ID 0 */
		0x10, 0x0b, 0x0c, 0x0e, 0x0c, 0x0a, 0x10, 0x0e,
		0x0d, 0x0e, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28,
		0x1a, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23, 0x25,
		0x1d, 0x28, 0x3a, 0x33, 0x3d, 0x3c, 0x39, 0x33,
		0x38, 0x37, 0x40, 0x48, 0x5c, 0x4e, 0x40, 0x44,
		0x57, 0x45, 0x37, 0x38, 0x50, 0x6d, 0x51, 0x57,
		0x5f, 0x62, 0x67, 0x68, 0x67, 0x3e, 0x4d, 0x71,
		0x79, 0x70, 0x64, 0x78, 0x5c, 0x65, 0x67, 0x63,
		0x01,		/* table ID 1 */
		0x11, 0x12, 0x12, 0x18, 0x15, 0x18, 0x2f, 0x1a,
		0x1a, 0x2f, 0x63, 0x42, 0x38, 0x42, 0x63, 0x63,
		0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
		0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
		0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
		0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
		0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
		0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63
	};

	/* write fixed quantitization tables */
	adr = 0x0cc;
	for (i = 0; i < sizeof(dqt); ++i)
		zr36060_write_8(btv, adr++, dqt[i]);
}

static void zr36060_set_jpg_DHT(struct buz * btv)
{
	unsigned i;
	unsigned adr;
	static const u8 dht[] = {
		0xff, 0xc4,	/* DHT marker */
		0x01, 0xa2,	/* DHT length */
		0x00,		/* table class 0, ID 0 */
		0x00, 0x01, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, /* # codes of length 1..8 */
		0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* # codes of length 8..16 */
		0x00,						/* values for codes of length 2 */
		0x01, 0x02, 0x03, 0x04, 0x05,			/* values for codes of length 3 */
		0x06,						/* values for codes of length 4 */
		0x07,						/* values for codes of length 5 */
		0x08,						/* values for codes of length 6 */
		0x09,						/* values for codes of length 7 */
		0x0a,						/* values for codes of length 8 */
		0x0b,						/* values for codes of length 9 */
		0x01,		/* table class 0, ID 1 */
		0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,	/* # codes of length 1..8 */
		0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,	/* # codes of length 9..16 */
		0x00, 0x01, 0x02,				/* values for codes of length 2 */
		0x03,						/* values for codes of length 3 */
		0x04,						/* values for codes of length 4 */
		0x05,						/* values for codes of length 5 */
		0x06,						/* values for codes of length 6 */
		0x07,						/* values for codes of length 7 */
		0x08,						/* values for codes of length 8 */
		0x09,						/* values for codes of length 9 */
		0x0a,						/* values for codes of length 10 */
		0x0b,						/* values for codes of length 11 */
		0x10,
		0x00, 0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03,
		0x05, 0x05, 0x04, 0x04, 0x00, 0x00, 0x01, 0x7d,
		0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
		0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
		0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
		0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
		0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
		0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
		0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
		0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
		0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
		0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
		0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
		0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
		0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
		0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
		0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
		0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
		0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
		0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
		0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
		0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
		0xf9, 0xfa, 0x11, 0x00, 0x02, 0x01, 0x02, 0x04,
		0x04, 0x03, 0x04, 0x07, 0x05, 0x04, 0x04, 0x00,
		0x01, 0x02, 0x77, 0x00, 0x01, 0x02, 0x03, 0x11,
		0x04, 0x05, 0x21, 0x31, 0x06, 0x12, 0x41, 0x51,
		0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81, 0x08,
		0x14, 0x42, 0x91, 0xa1, 0xb1, 0xc1, 0x09, 0x23,
		0x33, 0x52, 0xf0, 0x15, 0x62, 0x72, 0xd1, 0x0a,
		0x16, 0x24, 0x34, 0xe1, 0x25, 0xf1, 0x17, 0x18,
		0x19, 0x1a, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x35,
		0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44, 0x45,
		0x46, 0x47, 0x48, 0x49, 0x4a, 0x53, 0x54, 0x55,
		0x56, 0x57, 0x58, 0x59, 0x5a, 0x63, 0x64, 0x65,
		0x66, 0x67, 0x68, 0x69, 0x6a, 0x73, 0x74, 0x75,
		0x76, 0x77, 0x78, 0x79, 0x7a, 0x82, 0x83, 0x84,
		0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92, 0x93,
		0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xa2,
		0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa,
		0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
		0xba, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8,
		0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
		0xd8, 0xd9, 0xda, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6,
		0xe7, 0xe8, 0xe9, 0xea, 0xf2, 0xf3, 0xf4, 0xf5,
		0xf6, 0xf7, 0xf8, 0xf9, 0xfa
	};

	/* write fixed Huffman tables */
	adr = 0x1d4;
	for (i = 0; i < sizeof(dht); ++i)
		zr36060_write_8(btv, adr++, dht[i]);
}

static void zr36060_set_jpg_APP(struct buz * btv)
{
	unsigned i;
	unsigned adr;
	static const u8 app[] = {
		0xff, 0xe0,	/* APP marker */
		0x00, 0x3e,	/* APP length */
		0x41, 0x56, 0x49, 0x31, 0x01, 0x20, 0x20, 0x20,
		0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
		0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
		0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
		0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
		0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
		0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
		0x20, 0x20, 0x20, 0x20
	};

	/* write fixed app data */
	adr = 0x380;
	for (i = 0; i < sizeof(app); ++i)
		zr36060_write_8(btv, adr++, app[i]);
}

static void zr36060_set_jpg_COM(struct buz * btv)
{
	unsigned i;
	unsigned adr;
	static const u8 com[] = {
		0xff, 0xfe,	/* COM marker */
		0x00, 0x3e,	/* COM length */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00
	};

	/* write fixed com data */
	adr = 0x3c0;
	for (i = 0; i < sizeof(com); ++i)
		zr36060_write_8(btv, adr++, com[i]);
}

static void zr36060_set_cap(struct buz * btv)
{
	unsigned i;
	u32 reg;

	zr36060_reset(btv);
	mdelay(10);

	reg	= (0 << 7)	/* Load=0 */
		| (1 << 0);	/* SynRst=1 */
	zr36060_write_8(btv, 0x000, reg);

	zr36060_set_jpg(btv);
	zr36060_set_video(btv);
	zr36060_set_jpg_SOF(btv);
	zr36060_set_jpg_SOS(btv);
	zr36060_set_jpg_DRI(btv);
	zr36060_set_jpg_DQT(btv);
	zr36060_set_jpg_DHT(btv);
	zr36060_set_jpg_APP(btv);
	zr36060_set_jpg_COM(btv);

	reg	= (1 << 7)	/* Load=1 */
		| (0 << 0);	/* SynRst=1 */
	zr36060_write_8(btv, 0x000, reg);

	/* wait for codec to unbusy */
	for (i = 0; i < 1000; ++i)
	{
		reg = zr36060_read_8(btv, 0x001);
		if ((reg & (1 << 7)) == 0)
		{
			printk(KERN_INFO "060: loaded, loops=%u\n", i);
			//zr36060_dump(btv);
			return;
		}
		udelay(1000);
	}
	printk(KERN_INFO "060: stuck busy, statux=%02x\n", reg);
}

static void video_reset(struct buz * btv)
{
	GPIO(btv, 0, 0);
	mdelay(10);
	GPIO(btv, 0, 1);
	mdelay(10);
}

static int zr36057_init(int i)
{
        struct buz *btv = &buzs[i];
	u32 * mem;
	unsigned mem_needed;
	int j;

	/* reset zr36057 */
	btwrite(ZR36057_SPGPPCR_SoftReset, 0);
	mdelay(10);

        btv->picture.colour=254<<7;
        btv->picture.brightness=128<<8;
        btv->picture.hue=128<<8;
        btv->picture.contrast=0xd8<<7;

        btv->user=0; 

        btv->fbuffer=NULL;

        /* default setup for max. NTSC size in a 1024xXXX hicolor framebuffer */

        btv->win.norm=VIDEO_MODE_NTSC;
        btv->win.interlace=1;
        btv->win.x=0;
        btv->win.y=0;
        btv->win.width=tvnorms[btv->win.norm].Wa;
        btv->win.height=tvnorms[btv->win.norm].Ha;
        btv->win.cropwidth=tvnorms[btv->win.norm].Wa;
        btv->win.cropheight=tvnorms[btv->win.norm].Ha;
        btv->win.cropx=0;
        btv->win.cropy=0;
        btv->win.bpp=2;
        btv->win.color_fmt=VIDEO_PALETTE_RGB565;
        btv->win.swidth=1024;
        btv->win.bpl=btv->win.swidth*btv->win.bpp;
        btv->win.sheight=768;
        btv->cap=0;

        btv->gmode=0;
        btv->stat_com=0;
        btv->grab_head=0;
        btv->grab_tail=0;

	/* allocate memory *before* doing anything to the hardware in case allocation fails */
	mem_needed = (BUZ_NUM_FBUF)					/* for STAT_COM table */
		+ (BUZ_NUM_FBUF * BUZ_NUM_FRAGS * 2)			/* for FRAG_TABles */
		+ (((BUZ_MAX_WIDTH + 31) / 32) * BUZ_MAX_HEIGHT);	/* for overlay mask */
        if (!(mem=(u32 *) kmalloc(mem_needed * 4, GFP_KERNEL)))
	{
		rvfree((void *) btv->fbuffer, BUZ_NUM_FBUF*BUZ_MAX_FBUF);
		btv->fbuffer = NULL;
		return -ENOMEM;
	}
        memset(mem, 0, mem_needed * 4);

        printk(KERN_DEBUG "buz: stat_com=%p size=0x%x\n", mem, BUZ_NUM_FBUF*4);
	btv->stat_com = mem;
	mem += (BUZ_NUM_FBUF);
	for (j = 0; j < BUZ_NUM_FBUF; ++j)
	{
		btv->stat_com[j] = 1;		/* mark as unavailable to zr36057 */
		printk(KERN_DEBUG "buz: frag_tab[%d]=%p size=0x%lx\n", i, mem, BUZ_NUM_FRAGS*2*4);
		btv->frag_tab[j] = mem;
		btv->frag_tab_bus[j] = virt_to_bus(mem);
		mem += (BUZ_NUM_FRAGS * 2);
	}
        printk(KERN_DEBUG "buz: overlay_mask=%p size=0x%x\n", mem, ((BUZ_MAX_WIDTH + 31) / 32) * BUZ_MAX_HEIGHT * 4);
        btv->overlay_mask = mem;
	mem += (((BUZ_MAX_WIDTH + 31) / 32) * BUZ_MAX_HEIGHT);


        printk(KERN_DEBUG "buz: stat_com=%p overlay_mask=%p mem_needed=0x%x mem_used=0x%x\n",
		btv->stat_com, btv->overlay_mask, mem_needed*4, (u32)mem - (u32)btv->stat_com);
        printk(KERN_DEBUG "buz: frag_tab[0]=%p frag_tab[1]=%p frag_tab[2]=%p frag_tab[3]=%p\n",
		btv->frag_tab[0], btv->frag_tab[1], btv->frag_tab[2], btv->frag_tab[3]);

	/* take zr36057 out of reset now */
	btwrite(ZR36057_SPGPPCR_SoftReset, ZR36057_SPGPPCR);
	mdelay(10);

	/* stop all DMA processes */
	btwrite(ZR36057_MCTCR_CFlush, ZR36057_MCTCR);
	btor(ZR36057_VDCR_VidEn, ZR36057_VDCR);
	/* assert P_Reset */
	btwrite(0, ZR36057_JPC);

	/* set up GPIO direction */
	btwrite(ZR36057_SPGPPCR_SoftReset | 0, ZR36057_SPGPPCR);

	/* Set up guest bus timing - Guests 0..3 Tdur=12, Trec=3 */
	btwrite((gpio << 24) | 0x8888, ZR36057_GPPGCR1);
	mdelay(10);

	/* reset video decoder */
	video_reset(btv);

	/* reset JPEG codec */
	zr36060_sleep(btv, 0);
	mdelay(10);
	zr36060_reset(btv);
	mdelay(10);

        DEBUG(printk(KERN_DEBUG "buz: zr36057_mem: 0x%08x\n",(unsigned int) btv->zr36057_mem));

        /* i2c */
        memcpy(&btv->i2c, &buz_i2c_bus_template, sizeof(struct i2c_bus));
        sprintf(btv->i2c.name, "zr36057-%d", i);
        btv->i2c.data = btv;
        if (i2c_register_bus(&btv->i2c)<0)
	{
		kfree((void *) btv->stat_com);
		btv->stat_com = NULL;
                return -1;
	}

        //zr36057_set_winsize(btv);

        /*
         *	Now add the template and register the device unit.
         */
        memcpy(&btv->video_dev,&buz_template, sizeof(buz_template));
        if(video_register_device(&btv->video_dev,VFL_TYPE_GRABBER)<0)
	{
		i2c_unregister_bus(&btv->i2c);
		kfree((void *) btv->stat_com);
		btv->stat_com = NULL;
                return -1;
	}

	/* toggle JPEG codec sleep to sync PLL */
	zr36060_sleep(btv, 1);
	mdelay(10);
	zr36060_sleep(btv, 0);
	mdelay(10);

	/* Enable bus-mastering */
	pci_set_master(btv->pci_dev);

#if 0
	{
		int i, j;
		for (i = 0; i < 8; i++) {
			for (j = 0; j < 8; j++) {
				printk(KERN_INFO "060: %d,%d %02x\n", i, j, post_office_read(btv, i, j));
			}
		}
	}
#endif

//	zr36060_dump(btv);

	{
		i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_SET_INPUT, &input);
		i2c_control_device(&btv->i2c, I2C_DRIVERID_DECODER, DECODER_DUMP, NULL);
	}

	/* set individual interrupt enables but don't global enable until buz_open() */
	btwrite(ZR36057_ICR_GIRQ1 | ZR36057_ICR_GIRQ0 | ZR36057_ICR_JPEGRepIRQ, ZR36057_ICR);

        return 0;
}



/*
 *	Scan for a Bt848 card, request the irq and map the io memory 
 */

static int find_zr36057(void)
{
	unsigned char revision;
	unsigned char latency;
	struct buz *btv;
        struct pci_dev *dev = NULL;
	int result;
	int i;

	buz_num=0;

	while ((dev = pci_find_device(PCI_VENDOR_ID_ZORAN, PCI_DEVICE_ID_ZORAN_36057, dev)) != NULL)
	{
		btv=&buzs[buz_num];
		btv->pci_dev=dev;
		btv->zr36057_mem=NULL;
		for (i = 0; i < BUZ_NUM_FBUF; ++i)
			btv->stat_com_capq[i] = NULL;

                btv->zr36057_adr = btv->pci_dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK;
		pci_read_config_byte(btv->pci_dev, PCI_CLASS_REVISION, &revision);
		printk(KERN_INFO "buz: Zoran ZR36057 (rev %d) irq: %d, memory: 0x%08x.\n",
                        revision, btv->pci_dev->irq, btv->zr36057_adr);

		btv->zr36057_mem=ioremap(btv->zr36057_adr, 0x1000);

		result = request_irq(btv->pci_dev->irq, buz_irq,
			SA_SHIRQ | SA_INTERRUPT,"buz",(void *)btv);
		if (result < 0) 
		{
			if (result==-EINVAL) 
			{
				printk(KERN_ERR "buz: Bad irq number or handler\n");
			}
			if (result==-EBUSY)
			{
				printk(KERN_ERR "buz: IRQ %d busy, change your PnP config in BIOS\n",btv->pci_dev->irq);
			}
			iounmap(btv->zr36057_mem);
			return result;
		}

		/* set PCI latency timer */
		pci_read_config_byte(btv->pci_dev, PCI_LATENCY_TIMER, &latency);
		if (latency != 48)
		{
			printk(KERN_INFO "buz: Changing PCI latency from %d to 48.\n", latency);
			latency = 48;
			pci_write_config_byte(btv->pci_dev, PCI_LATENCY_TIMER, latency);
		}

		buz_num++;
	}
	if(buz_num)
		printk(KERN_INFO "buz: %d Buz card%s found.\n", buz_num, buz_num == 1?"":"s");
	find_vga();
	return buz_num;
}


static void release_buz(void)
{
	u8 command;
	int i;
	struct buz *btv;

	for (i=0;i<buz_num; i++) 
	{
		btv=&buzs[i];

    		/* unregister i2c_bus */
		i2c_unregister_bus((&btv->i2c));

		/* disable PCI bus-mastering */
		pci_read_config_byte(btv->pci_dev, PCI_COMMAND, &command);
		command&=~PCI_COMMAND_MASTER;
		pci_write_config_byte(btv->pci_dev, PCI_COMMAND, command);

		/* put chip into reset */
		btwrite(0, ZR36057_SPGPPCR);

		free_irq(btv->pci_dev->irq,btv);

		/* unmap and free memory */
		if (btv->stat_com)
		{
			kfree((void *) btv->stat_com);
			btv->stat_com = NULL;
		}

		DEBUG(printk(KERN_DEBUG "zr36057_mem: 0x%p.\n", btv->zr36057_mem));
		if (btv->zr36057_mem)
		{
			iounmap(btv->zr36057_mem);
			btv->zr36057_mem = 0;
		}

		video_unregister_device(&btv->video_dev);
	}
}

#ifdef MODULE

int init_module(void)
{
#else
int init_buz_cards(struct video_init *unused)
{
#endif
	int i;

	handle_chipset();
	if (find_zr36057()<0)
		return -EIO;

	/* initialize Bt848s */
	for (i=0; i<buz_num; i++) 
	{
		if (zr36057_init(i)<0) 
		{
			release_buz();
			return -EIO;
		} 
	}
	return 0;
}



#ifdef MODULE

void cleanup_module(void)
{
        release_buz();
}

#endif

/*
 * Local variables:
 * c-indent-level: 8
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -8
 * c-argdecl-indent: 8
 * c-label-offset: -8
 * c-continued-statement-offset: 8
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
