/* 
    buz - Iomega Buz frame grabber driver

    Copyright (C) 1998 Dave Perks <dperks@ibm.net>

	based on

    bttv - Bt848 frame grabber driver

    Copyright (C) 1996,97 Ralph Metzler (rjkm@thp.uni-koeln.de)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _BUZ_H_
#define _BUZ_H_

#define TEST_VBI

#include <linux/types.h>
#include <linux/wait.h>

#include "/usr/src/linux/drivers/char/i2c.h"
#include "zr36057.h"
#include <linux/videodev.h>

#define MAX_CLIPRECS	100
#define BUZ_MAX_FBUF	(100 * 1024)
#define BUZ_NUM_FBUF	4
#define BUZ_MASK_FBUF	3
#define	BUZ_NUM_FRAGS	(BUZ_MAX_FBUF / PAGE_SIZE + 2)


/* clipping rectangle */
struct cliprec 
{
	int x, y, x2, y2;
	struct cliprec *next;
};


/* grab buffer */
struct gbuffer 
{
	void *adr;
	int x, y;
	int width, height;
	unsigned int bpl;
	unsigned int fmt;
	int flags;
#define GBUF_ODD  1
#define GBUF_EVEN 2
#define GBUF_LFB  4
#define GBUF_INT  8
	unsigned int length;
	void *ro;
	void *re;
	u32 bro;
	u32 bre;
};


#ifdef __KERNEL__

struct buz_window 
{
	int x, y;
	ushort width, height;
	ushort bpp, bpl;
	ushort swidth, sheight;
	short cropx, cropy;
	ushort cropwidth, cropheight;
	u32 vidadr;
	ushort freq;
	int norm;
	int interlace;
	int color_fmt;
	ushort depth;
};


struct buz 
{
	struct video_device video_dev;
	struct i2c_bus i2c;

	struct video_picture picture;		/* Current picture params */
	struct buz_window win;
	int got_window;
	int got_buffer;
	u32 *overlay_mask;

	unsigned short id;
	struct pci_dev *pci_dev;          /* PCI device */
	unsigned int zr36057_adr;      /* bus address of IO mem returned by PCI BIOS */
	unsigned char *zr36057_mem;   /* pointer to mapped IO memory */
  
	int user;

	int cap;
	struct cliprec *cliprecs;
	int ncr;		/* number of clipping rectangles */

	u16 gwidth, gheight, gfmt;
	int gmode;

	/* grab queue counts/indices, mask with BUZ_MASK_FBUF before using as index */
	/* (grab_tail - grab_head) is number active in DMA, must be <= BUZ_NUM_FBUF */
	/* (value & BUZ_MASK_FBUF) corresponds to index in stat_com table */
	unsigned grab_head;			/* first grab buffer in active queue */
	unsigned grab_tail;			/* first grab buffer after queue */

	/* zr36057's code buffer table */
	u32 * stat_com;				/* stat_com[i] is indexed by grag_head/tail & BUZ_MASK_FBUF */
	int stat_com_frame[BUZ_NUM_FBUF]; /* this too, indicates which frag buffer is here */
	int stat_com_result[BUZ_NUM_FBUF];	/* result returned by CSYNC call */
	struct wait_queue * stat_com_capq[BUZ_NUM_FBUF]; /* wait here for grab to finish */
	int stat_com_busy[BUZ_NUM_FBUF];

	/* arrays are indexed by frame number */
	char * fbuffer;
	u32 * frag_tab[BUZ_NUM_FBUF];		/* addresses of frag tables */
	u32 frag_tab_bus[BUZ_NUM_FBUF];		/* same values cached to save time in ISR */
	int frag_tab_busy[BUZ_NUM_FBUF];	/* non-zero if corresponding buffer is in use in grab queue */
	int frag_tab_index[BUZ_NUM_FBUF];	/* index into stat_com[] */
};

#endif

/*The following should be done in more portable way. It depends on define
  of _ALPHA_BUZ in the Makefile.*/

#ifdef _ALPHA_BUZ
#define btwrite(dat,adr)    writel((dat),(char *) (btv->zr36057_adr+(adr)))
#define btread(adr)         readl(btv->zr36057_adr+(adr))
#else
#define btwrite(dat,adr)    writel((dat), (char *) (btv->zr36057_mem+(adr)))
#define btread(adr)         readl(btv->zr36057_mem+(adr))
#endif

#define btand(dat,adr)      btwrite((dat) & btread(adr), adr)
#define btor(dat,adr)       btwrite((dat) | btread(adr), adr)
#define btaor(dat,mask,adr) btwrite((dat) | ((mask) & btread(adr)), adr)

#define I2C_TSA5522        0xc2
#define I2C_TDA9850        0xb6
#define I2C_HAUPEE         0xa0
#define I2C_STBEE          0xae
#define	I2C_SAA7111        0x48
#define	I2C_SAA7185        0x88

#define TDA9850_CON1       0x04
#define TDA9850_CON2       0x05
#define TDA9850_CON3       0x06
#define TDA9850_CON4       0x07
#define TDA9850_ALI1       0x08
#define TDA9850_ALI2       0x09
#define TDA9850_ALI3       0x0a

#endif
